/********************************************************************************/
/*    H   H Y   Y TTTTT EEEEE  CCC       HYTEC ELECTRONICS LTD                  */
/*    H   H  Y Y    T   E     C          5 Cradock Road,                        */
/*    HHHHH   Y     T   EEE   C          Reading, Berks.    Tel: 0118 9757770   */
/*    H   H   Y     T   E     C          RG2 0JT            Fax: 0118 9757566   */
/*    H   H   Y     T   EEEEE  CCC       Web: www.hytec-electronics.co.uk       */
/********************************************************************************/

/********************************************************************************/
/*                                                                              */
/*  drvHy8505.c - Driver Support for the Hytec 8505IP cards                     */
/*                                                                              */
/*                               .                                              */
/*                                                                              */
/********************************************************************************/
/*                                                                              */
/*  Revision history: (comment and initial revisions)                           */
/*                                                                              */
/*  vers.       revised         modified by                                     */
/*  -----       -----------     -------------------------                       */
/*  1.1         01-JUL-2005     Paul Hamadyk DLS Ltd                            */
/*	                        Driver support for a Hytec Hy8505 IP            	*/
/*			        module. Initial Version.                        			*/
/*  1.2         01-NOV-2010     Jim Chen Hytec Electronics Ltd                  */
/*	                        Device support for a Hytec Hy8505 IP            	*/
/*			        		module. Modified to osi compliant and         		*/
/*                          took out the Hotswap support.                       */
/*                                                                              */
/********************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <devLib.h>
#include <dbScan.h>
#include <drvSup.h>

#include "Hy8505.h"
#include <drvIpac.h>

#include <epicsExit.h>
#include <registryFunction.h>
#include <iocsh.h>
#include <epicsExport.h>
#include <dbDefs.h>

#undef DEBUG

/* io structure */

struct ioregs
{

/* writable I/O registers */

	unsigned short LKC;	/* last known change */
	unsigned short CSR;	/* control / status register */
	unsigned short IMR;	/* interrupt mask register */
	unsigned short DBR;	/* debounce register */
	unsigned short PSR;	/* pulse select register */
	unsigned short PPR;	/* pulse parameter register */
	unsigned short VEC;	/* interrupt vector */
};

/* structure to support multiple IP modules */

struct cardinfo
{

/* next structure */

	struct cardinfo *next;

/* for EPICS I/O interrupt */

	IOSCANPVT ioscanpvt;

/* config details */
	int invertInputs;			/* If set all input bits are inverted. */
	int invertOutputs;			/* If set all output bits are inverted. */
	
	unsigned short cardnum;		/* EPICS card number */
	unsigned short carrier;		/* carrier card number */
	unsigned short ipslot;		/* IP slot number */
	unsigned short init;		/* module initialised */
	unsigned short intr;		/* IP interrupt vector */
	unsigned short intmask;		/* interrupt mask */
	unsigned short state;		/* insertion state */
	unsigned short debrate;		/* debounce rate */
	
	struct ioregs *regs;		/* registers in I/O space on card */
	struct ioregs restore;		/* registers to restore */
};

static long REPORT(int level);
static long INIT(void);
static long LIST(void);

static int dumpcard(struct cardinfo *card);

long drvHy8505_read (const int cardnum, unsigned short *value);
long drvHy8505_write(const int cardnum, const unsigned short mask, unsigned short value);

/* our single linked list of cardinfo structures */

static struct cardinfo *cardlist = NULL;

struct drv_sup
{
	long		number;
	DRVSUPFUN	report;
	DRVSUPFUN	init;
};

struct drv_sup drvHy8505 =
{
	2,
	REPORT,
	INIT
};

static long drvHy8505Report(int level)
{

	switch (level)
	{
		case 0:		;
		case 1:		LIST();
				break;
		case 2:
		{
			struct cardinfo *card = cardlist;
			
		/* display the registers */
		
			while (card != NULL)
			{		
				printf("\tLKC: %04x\tCSR: %04x\n", card->regs->LKC, card->regs->CSR);
				printf("\tIMR: %04x\tDBR: %04x\n", card->regs->IMR, card->regs->DBR);
				printf("\tPSR: %04x\tPPR: %04x\n", card->regs->PSR, card->regs->PPR);
				printf("\tVEC: %04x\n", card->regs->VEC);
				
				card = card->next;
			}
			
			break;
		}
	
		default:
		{
			break;
		}
	}
	
	return (0);
}


/*
	device layer masks out all bits defined by the signal number and the shift
*/

long drvHy8505_read(const int cardnum, unsigned short *value)
{
	struct cardinfo *card = cardlist;
	
	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		return (S_drv_badParam);
	}

/* read inputs, invert if required */
	
	*value = card->regs->LKC;
	if (card->invertInputs)
		*value = ~ *value;

#ifdef DEBUG

	printf("drvHy8505_read - card %2d  val 0x%04x\n", cardnum, *value);
	
#endif	
	
	return (S_drv_OK);
}

/*
	Set individual bits

*/

long drvHy8505_write(const int cardnum, const unsigned short mask, unsigned short value)
{
	struct cardinfo *card = cardlist;

	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		return (S_drv_badParam);
	}

/* write outputs, inverting if requested */

	if (card->invertOutputs)
		value = ~value;
	if (mask == 0xffff)
	{
		card->regs->LKC = card->restore.LKC = value;
	}
	else
	{
		card->restore.LKC = card->regs->LKC = (card->regs->LKC & ~mask) | (value & mask);
	}
	
#ifdef DEBUG

	printf("drvHy8505_write - card %2d  mask 0x%04x  val 0x%04x\n", cardnum, mask, value);
	
#endif
	
	return (S_drv_OK);
}

/*********************************************************************/

#define PR_ID    0x06
#define PR_MODN  0x0a
#define PR_REVN  0x0c
#define PR_DRID1 0x10
#define PR_DRID2 0x12
#define PR_FLAGS 0x14
#define PR_BYTES 0x16
#define PR_SERN  0x1a
#define HYTECID  0x00800300

/*
	check that the correct IP module is installed
*/

static int checkprom(char *ptr)
{
	int ret = 0;
	unsigned int manid;
	unsigned short model, rev, serial;
	char header[16], tmp;
	
	if(strcmp("VITA4 ", ptr) != 0) 
		ret += 1;
	memcpy(header, ptr, 2);
	tmp = header[1];
	header[1] = header[0];
	header[0] = tmp;
	memcpy(&header[2], ptr+2, 2);
	tmp = header[3];
	header[3] = header[2];
	header[2] = tmp;	
	memcpy(&header[4], ptr+4, 2);
	header[6]=0;

	manid  = *((short *)   (ptr +  8))  + ((*((short *)(ptr +  6))) << 16);
	model  = *((short *) (ptr + 10));
	rev    = *((short *) (ptr + 12));
	serial = *((short *) (ptr + 26));

	printf("%s: %s  model: %04X  manid: %08X  rev: %04X  serial: %d\n", DRIVER, header, model, manid, rev, serial);

	if (manid != MANID) ret += 2;
	if (model != MODEL) ret += 4;
	if (rev    < REV)   ret += 8;

	return (ret);
}

/*
	write data back to the IP module
	
	The order is important: CSR should be written before LKC so that the card is initialised for
	input or output before the output is set.
*/

static int dumpcard(struct cardinfo *card)
{
/*	card->regs->LKC = card->restore.LKC;	*/		/* we don't need to do this. imagine it is input mode */
	card->regs->IMR = card->restore.IMR;
	card->regs->DBR = card->restore.DBR;
	card->regs->PSR = card->restore.PSR;
	card->regs->PPR = card->restore.PPR;
	card->regs->VEC = card->restore.VEC;

	card->regs->CSR = card->restore.CSR & 0xBFFF;	/* disable interrupt at start */
	card->regs->CSR = card->restore.CSR & 0x3FFF;	/* clear interrupt as well */
	card->regs->CSR = card->restore.CSR;			/* enable interrupt */

	return OK;
}

/*
	interrupt service routine
*/

static void drvISR(const int cardnum)
{
/*printf("====8505 int\n");*/
	struct cardinfo *card = cardlist;
	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		return;
	}
	
/* clear the interrupt */

	card->regs->CSR &= 0x7fff;
/* inform EPICS */
	scanIoRequest(card->ioscanpvt);

/* Note, the interrupt enable might need to be done in ioscanpvt */
	ipmIrqCmd(card->carrier, card->ipslot, 0, ipac_irqEnable);	
}

/*
	list the cards and ipslots
*/

static long LIST(void)
{
	struct cardinfo *card = cardlist;

	while (card != NULL)
	{
		printf("%s:  card: %d  carrier: %2d  ip: %d  %s\n", DRIVER, card->cardnum, card->carrier, card->ipslot, (card->init) ? "configured" : "");
		card = card->next;
	}
	
	return (0);
}

/*
	Checks that the parameters are correct and that the card hasn't already been configured.
	EPICS will call the driver initialisation function later which will do the real configuration.
	
	cardnum		EPICS card number (DLS Ltd = VME slot * 10 + IP slot)
	carrier		carrier number returned by ipacEXTAddCarrier
	ipslot		ipslot 0-3
	debrate		debounce rate	0 = none
					1 = 100Hz
					2 = 200Hz
					3 = 500Hz
					4 = 1kHz

	debmask		select input bits to debounce 0x0000 - 0xFFFF
	
	scanrate	input scan rate	0 = 1kHz
					1 = 10kHz
					2 = 100kHz
					3 = 1MHz
	dir		0 = inputs
			1 = low outputs / high inputs
			2 = low inputs / high outputs
			3 outputs

			Bits 2 & 3 (values 4 and 8, respectively) encode whether the input
			and output values should be inverted.
			(dir & 4) == 4  =>  Invert all input bits
			(dir & 8) == 8  =>  Invert all output bits
			
	intr		interrupt vector
	
	intmask		select bits to generate interrupts	0x0000 - 0xFFFF

	pwidth		pulse width	0 = 1 msec
					1 = 10 msec
					2 = 100 msec
					3 = 1 sec
					4 = 2 sec
					5 = 5 sec
					6 = 10 sec
					7 = 20 sec
					8 = 50 sec
					9 = 100 sec

	pmask		select bits to pulse on output		0x0000 - 0xFFFF
	
	clock		0 = internal, 1 = external

*/

#define INVERT_INPUT_BIT	4
#define INVERT_OUTPUT_BIT	8


int CONFIGURE(	const int cardnum, const int carrier, const int ipslot,
		const int debrate, const int pwidth, const int scanrate, 
		const int dir, const int intr, const int clock)
{
	struct cardinfo *card = cardlist;
	const char *dirs[] = { "in/in", "out/in", "in/out", "out/out"};
	const char *scans[] = { "1k", "10k", "100k", "1M"};
	const char *bounce[] = { "100", "200", "500", "1k"};
	const char *widths[] = { "1 msec", "10 msec", "100 msec", "1 sec", "2 sec", "5 sec", "10 sec", "20 sec", "50 sec", "100 sec"};

/* check parameters */

	while (card != NULL)
	{
		if (card->cardnum == cardnum) 
		{
			printf("%s: card %d already defined.\n", DRIVER, cardnum);
			card->init = 0xffff;
			return (-1);
		}
		else
		{
			if ((card->carrier == carrier) && (card->ipslot == ipslot))
			{
				printf("%s: Carrier and IP slot already defined.\n", DRIVER);
				card->init = 0xffff;
				return (-1);
			}
		}
			
		card = card->next;
	}

/* check parameters */

	if ((debrate < 0) || (debrate > 4))
	{
		printf("%s: debounce rate parameter %d invalid\n", DRIVER, debrate);
		card->init = 0xffff;
		return (-1);
	}

	if ((scanrate < 0) || (scanrate > 3))
	{
		printf("%s: scan rate parameter %d invalid\n", DRIVER, scanrate);
		card->init = 0xffff;
		return (-1);
	}

	if ((dir < 0) || (dir > 15))
	{
		printf("%s: direction parameter %d invalid \n", DRIVER, dir);
		card->init = 0xffff;
		return (-1);
	}

	if ((pwidth < 0) || (pwidth > 9))
	{
		printf("%s: pulse width parameter %d invalid\n", DRIVER, pwidth);
		card->init = 0xffff;
		return (-1);
	}

	if ((clock != 0) && (clock != 1))
	{
		printf("%s: clock parameter %d invalid\n", DRIVER, clock);
		card->init = 0xffff;
		return (-1);
	}
	
/* allocate structure and initialise */

	card = calloc(1, sizeof(struct cardinfo));
	
	if (card == NULL)
	{
		printf("%s: Memory allocation error.\n", DRIVER);
		card->init = 0xffff;
		return (-1);
	}

	card->next = cardlist;
	cardlist   = card;

	card->invertInputs  = (dir & INVERT_INPUT_BIT) != 0;
	card->invertOutputs = (dir & INVERT_OUTPUT_BIT) != 0;
	card->cardnum = cardnum;
	card->carrier = carrier;
	card->ipslot  = ipslot;
	card->intr    = intr & 0xff;		/* interrupt vector */
	card->intmask = 0xffff;				/* initialise all interrupt mask */
	
	card->restore.LKC = 0x0000;
	card->restore.CSR = CSR_ENABLE;		/* enable input scanning */
	
/* debounce enable */

	if (debrate > 0)
	{
		card->debrate = debrate;
			
		card->restore.CSR |= (debrate - 1) << DEBOUNCE_SHIFT;
		card->restore.DBR  = 0xffff;
	}
	
	card->restore.CSR |= scanrate << SCAN_SHIFT;
	card->restore.CSR |= (dir & 3) << DIRECTION_SHIFT;
	card->restore.CSR |= (clock & 0x3) << CLOCK_SHIFT;
	
	card->restore.PSR = 0x0000;
	card->restore.PPR = pwidth;
				
	printf("%s:  card: %d  carrier: %2d  ip: %d  dir: %s\n%s:  scan: %sHz  debounce: %sHz  pulse: %s  int: %d\n", DRIVER, cardnum, carrier, ipslot, dirs[dir], DRIVER, scans[scanrate], bounce[debrate - 1], widths[pwidth], card->intr);
		
	return (0);
}

int Hy8505ExtraConfig(const int cardnum, const int debmask, const int pmask, const int intmask)
{
	struct cardinfo *card = cardlist;

	while ((card != NULL) && (card->cardnum != cardnum))
	{
		card = card->next;
	}
	
/* card wasn't found */

	if (card == NULL)
	{
		printf("%s:   card number %d not found\n", DRIVER, cardnum);
		card->init = 0xffff;
		return (-1);
	}
	
	if (intmask > 0xffff)
	{
		printf("%s: interrupt mask 0x%x invalid\n", DRIVER, intmask);
		card->init = 0xffff;
		return (-1);
	}

	if (debmask > 0xffff)
	{
		printf("%s: debounce mask 0x%x invalid\n", DRIVER, debmask);
		card->init = 0xffff;
		return (-1);
	}

	if (pmask > 0xffff)
	{
		printf("%s: pulse select mask 0x%x invalid\n", DRIVER, pmask);
		card->init = 0xffff;
		return (-1);
	}

	if (card->debrate > 0)
	{
		card->restore.DBR = debmask;
	}

	card->restore.PSR = pmask;
	card->intmask = intmask;

	printf("%s:  intmask:%x, pmask:%x, debmask:%x \n", DRIVER, intmask, pmask, debmask);
	
	return (0);
}

static long drvHy8505Init(void)
{
	struct cardinfo *card = cardlist;
	int status;

	while (card != NULL)
	{

	/* Don't continue if there was a fault with the configuration */

		if (card->init == 0xffff)
		{
			printf("%s: Error in configuration!\n", DRIVER);
			return (-1);
		}

	/* register the card in A16 address space */
	
		card->regs = (struct ioregs *) ipmBaseAddr(card->carrier, card->ipslot, ipac_addrIO);

		if (card->regs == NULL)
		{
			printf("%s: ipmBaseAddr failed for card %d.\n", DRIVER, card->cardnum);
			return (-1);
		}
			
	/* reboot handler */

		short reset = 0x0000;			
		card->regs->CSR  = reset;
				
	/* check the ROM */
			
		status = checkprom((char *)(card->regs) + PROM_OFFS);
	
	/* older versions have limited pulse width */
	
		if (*(short *)((char *)(card->regs) + PROM_OFFS + 12) <= 0x3301)
		{
			if (card->restore.PPR > 3)
			{
				printf("%s: Error! Pulse width maximum is 1 sec for this version Hy8505.\n", DRIVER);	
				
				return (-1);
			}
		}
		
	/* init interrupts */

		if (card->intr > 0)
		{
			scanIoInit(&(card->ioscanpvt));
						
			if (ipmIntConnect(card->carrier, card->ipslot, card->intr, drvISR, card->cardnum) != 0)
			{
				printf("%s: ipmIntConnect failed.\n", DRIVER);
				return (-1);
			}
			else
			{
				card->restore.VEC  = card->intr;
				card->restore.IMR  = card->intmask;
				card->restore.CSR |= 0x4000;		/* IO Int Enable */
	
				ipmIrqCmd(card->carrier, card->ipslot, 0, ipac_irqEnable);
			}
		}	

		if (dumpcard(card) == 0)
		{											
			card->init = 1;
			
			printf("%s: card %d configured at ipmBaseAddr 0x%x\n", DRIVER, card->cardnum, (unsigned int) card->regs);
		}

		card = card->next;
	}
	
	return (0);
}

/*
	reset the cards before reboot
*/


/******************* Return pointer to ioscanpvt   *****************/

int Hy8505getioscanpvt(const int cardnum, IOSCANPVT *scanpvt)
{
	struct cardinfo *card = cardlist;
	
	while ((card != NULL) && (card->cardnum != cardnum)) card = card->next;
	
	if (card == NULL)
	{
		printf("%s: getioscanpvt: cardnum %d not found!\n", DRIVER, cardnum);
		return (-1);
	}

	*scanpvt = card->ioscanpvt;
	
	return (0);
}


 epicsExportAddress(drvet, drvHy8505);

/* These two are for EPICS R3.13
 epicsRegisterFunction(Hy8505Configure);
 epicsRegisterFunction(Hy8505ExtraConfig);*/

 epicsRegisterFunction(Hy8505getioscanpvt);

/* Code for iocsh registration */

/* Hy8505Configure */
static const iocshArg Hy8505InitArg0 = {"cardNum", iocshArgInt};
static const iocshArg Hy8505InitArg1 = {"carrier", iocshArgInt};
static const iocshArg Hy8505InitArg2 = {"ipSlotNum", iocshArgInt};
static const iocshArg Hy8505InitArg3 = {"Debounce rate", iocshArgInt};
static const iocshArg Hy8505InitArg4 = {"Pulse With", iocshArgInt};
static const iocshArg Hy8505InitArg5 = {"Scan rate", iocshArgInt};
static const iocshArg Hy8505InitArg6 = {"Direction", iocshArgInt};
static const iocshArg Hy8505InitArg7 = {"Interrupt", iocshArgInt};
static const iocshArg Hy8505InitArg8 = {"clock rate", iocshArgInt};

static const iocshArg * const Hy8505InitArgs[] =  {&Hy8505InitArg0,
                                                       &Hy8505InitArg1,
                                                       &Hy8505InitArg2,
                                                       &Hy8505InitArg3,
                                                       &Hy8505InitArg4,
                                                       &Hy8505InitArg5,
                                                       &Hy8505InitArg6,
                                                       &Hy8505InitArg7,
                                                       &Hy8505InitArg8};

static const iocshFuncDef initHy8505 = {"Hy8505Configure", 9, Hy8505InitArgs};
static void initHy8505CallFunc(const iocshArgBuf *args)
{
    Hy8505Configure(args[0].ival, args[1].ival, args[2].ival, args[3].ival, args[4].ival,
                   args[5].ival, args[6].ival, args[7].ival, args[8].ival);
}


/* Hy8505ExtraConfig */

static const iocshArg Hy8505ExtraArg0 = {"cardNum", iocshArgInt};
static const iocshArg Hy8505ExtraArg1 = {"Debounce mask", iocshArgInt};
static const iocshArg Hy8505ExtraArg2 = {"Pulse mask", iocshArgInt};
static const iocshArg Hy8505ExtraArg3 = {"Interrupt mask", iocshArgInt};

static const iocshArg * const Hy8505ExtraArgs[] =  {&Hy8505ExtraArg0,
                                                       &Hy8505ExtraArg1,
                                                       &Hy8505ExtraArg2,
                                                       &Hy8505ExtraArg3};

static const iocshFuncDef configHy8505 = {"Hy8505ExtraConfig", 4, Hy8505ExtraArgs};
static void configHy8505CallFunc(const iocshArgBuf *args)
{
    Hy8505ExtraConfig(args[0].sval, args[1].ival, args[2].ival, args[3].ival);
}

static void Hy8505Register(void)
{
    iocshRegister(&initHy8505, initHy8505CallFunc);
    iocshRegister(&configHy8505, configHy8505CallFunc);
}

epicsExportRegistrar(Hy8505Register);

